@@ -1,6 +1,6 @@
 /* 
    HTTP utility functions
-   Copyright (C) 1999-2002, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 1999-2006, Joe Orton <joe@manyfish.co.uk>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -30,25 +30,29 @@
 #include <stdio.h>
 #include <ctype.h> /* isdigit() for ne_parse_statusline */
 
-#include "ne_utils.h"
-#include "ne_string.h" /* for ne_strdup */
-#include "ne_dates.h"
+#ifdef NE_HAVE_ZLIB
+#include <zlib.h>
+#endif
 
-#ifdef NEON_SSL
+#ifdef HAVE_OPENSSL
 #include <openssl/opensslv.h>
 #endif
 
-/* libxml2: pick up the version string. */
-#if defined(HAVE_XMLVERSION_H)
-#include <xmlversion.h>
-#elif defined(HAVE_LIBXML_XMLVERSION_H)
-#include <libxml/xmlversion.h>
+#ifdef HAVE_GNUTLS
+#include <gnutls/gnutls.h>
 #endif
 
-#ifdef NEON_ZLIB
-#include <zlib.h>
+/* libxml2: pick up the version string. */
+#if defined(HAVE_LIBXML)
+#include <libxml/xmlversion.h>
+#elif defined(HAVE_EXPAT) && !defined(HAVE_XMLPARSE_H)
+#include <expat.h>
 #endif
 
+#include "ne_utils.h"
+#include "ne_string.h" /* for ne_strdup */
+#include "ne_dates.h"
+
 int ne_debug_mask = 0;
 FILE *ne_debug_stream = NULL;
 
@@ -59,54 +63,76 @@ void ne_debug_init(FILE *stream, int mask)
 #if defined(HAVE_SETVBUF) && defined(_IONBF)
     /* If possible, turn off buffering on the debug log.  this is very
      * helpful if debugging segfaults. */
-    setvbuf(stream, NULL, _IONBF, 0);
+    if (stream) setvbuf(stream, NULL, _IONBF, 0);
 #endif        
 }
 
 void ne_debug(int ch, const char *template, ...) 
 {
     va_list params;
-    if ((ch&ne_debug_mask) != ch) return;
+    if ((ch & ne_debug_mask) == 0) return;
     fflush(stdout);
     va_start(params, template);
     vfprintf(ne_debug_stream, template, params);
     va_end(params);
-    if ((ch&NE_DBG_FLUSH) == NE_DBG_FLUSH) {
+    if ((ch & NE_DBG_FLUSH) == NE_DBG_FLUSH)
 	fflush(ne_debug_stream);
-    }
 }
 
-static const char *version_string = "neon " NEON_VERSION ": " 
+#define NE_STRINGIFY(x) # x
+#define NE_EXPAT_VER(x,y,z) NE_STRINGIFY(x) "." NE_STRINGIFY(y) "." NE_STRINGIFY(z)
+
+static const char version_string[] = "neon " NEON_VERSION ": " 
 #ifdef NEON_IS_LIBRARY
   "Library build"
 #else
   "Bundled build"
 #endif
+#ifdef NE_HAVE_IPV6
+   ", IPv6"
+#endif
 #ifdef HAVE_EXPAT
   ", Expat"
+/* expat >=1.95.2 exported the version */
+#ifdef XML_MAJOR_VERSION
+" " NE_EXPAT_VER(XML_MAJOR_VERSION, XML_MINOR_VERSION, XML_MICRO_VERSION)
+#endif
 #else /* !HAVE_EXPAT */
 #ifdef HAVE_LIBXML
-  ", libxml"
-#ifdef LIBXML_DOTTED_VERSION
-  " " LIBXML_DOTTED_VERSION
-#endif /* LIBXML_DOTTED_VERSION */
+  ", libxml " LIBXML_DOTTED_VERSION
 #endif /* HAVE_LIBXML */
 #endif /* !HAVE_EXPAT */
-#if defined(NEON_ZLIB) && defined(ZLIB_VERSION)
+#if defined(NE_HAVE_ZLIB) && defined(ZLIB_VERSION)
   ", zlib " ZLIB_VERSION
-#endif /* NEON_ZLIB && ... */
-#ifdef NEON_SOCKS
-   ", SOCKSv5 enabled"
+#endif /* NE_HAVE_ZLIB && ... */
+#ifdef NE_HAVE_SOCKS
+   ", SOCKSv5"
 #endif
-#ifdef NEON_SSL
+#ifdef NE_HAVE_LFS
+   ", LFS"
+#endif
+#ifdef HAVE_OPENSSL
 #ifdef OPENSSL_VERSION_TEXT
     ", " OPENSSL_VERSION_TEXT
 #else
    "OpenSSL (unknown version)"
 #endif /* OPENSSL_VERSION_TEXT */
-#else /* !NEON_SSL */
-   ", no OpenSSL support"
-#endif /* NEON_SSL */
+#ifdef NE_HAVE_TS_SSL
+    " (thread-safe)"
+#endif
+#endif /* HAVE_OPENSSL */
+#ifdef HAVE_GNUTLS
+    ", GNU TLS " LIBGNUTLS_VERSION
+#endif /* HAVE_GNUTLS */
+#ifdef HAVE_SSPI
+    ", SSPI"
+#endif /* HAVE_SSPI */
+#ifdef HAVE_PAKCHOIS
+    ", PKCS#11"
+#endif
+#ifdef NE_HAVE_LIBPXY
+    ", libproxy"
+#endif
    "."
 ;
 
@@ -115,29 +141,50 @@ const char *ne_version_string(void)
     return version_string;
 }
 
+#define LAST_COMPAT_ZERO_MINOR (27)
+
 int ne_version_match(int major, int minor)
 {
-    return (NEON_VERSION_MAJOR != major) || (NEON_VERSION_MINOR < minor);
+    return !
+        (NE_VERSION_MAJOR == 0 &&
+         (minor <= NE_VERSION_MINOR && minor >= LAST_COMPAT_ZERO_MINOR));
 }
 
-int ne_supports_ssl(void)
+int ne_has_support(int feature)
 {
-#ifdef NEON_SSL
-    return 1;
-#else
-    return 0;
+    switch (feature) {
+#if defined(NE_HAVE_SSL) || defined(NE_HAVE_ZLIB) || defined(NE_HAVE_IPV6) \
+    || defined(NE_HAVE_SOCKS) || defined(NE_HAVE_LFS) \
+    || defined(NE_HAVE_TS_SSL) || defined(NE_HAVE_I18N) || defined(HAVE_SSPI)
+#ifdef NE_HAVE_SSL
+    case NE_FEATURE_SSL:
 #endif
-}
-
-/* Cleanse 'str' of non-printable characters.  Duplicated in
- * ne_session.c for the duration of neon 0.23.x to prevent ABI
- * change. */
-static char *cleanse(char *str)
-{
-    char *pnt;
-    for (pnt = str; *pnt; pnt++)
-        if (iscntrl(*pnt) || !isprint(*pnt)) *pnt = ' ';
-    return str;
+#ifdef NE_HAVE_ZLIB
+    case NE_FEATURE_ZLIB:
+#endif
+#ifdef NE_HAVE_IPV6
+    case NE_FEATURE_IPV6:
+#endif
+#ifdef NE_HAVE_SOCKS
+    case NE_FEATURE_SOCKS:
+#endif
+#ifdef NE_HAVE_LFS
+    case NE_FEATURE_LFS:
+#endif
+#ifdef NE_HAVE_TS_SSL
+    case NE_FEATURE_TS_SSL:
+#endif
+#ifdef NE_HAVE_I18N
+    case NE_FEATURE_I18N:
+#endif
+#ifdef HAVE_SSPI
+    case NE_FEATURE_SSPI:
+#endif
+        return 1;
+#endif /* NE_HAVE_* */
+    default:
+        return 0;
+    }
 }
 
 int ne_parse_statusline(const char *status_line, ne_status *st)
@@ -145,41 +192,33 @@ int ne_parse_statusline(const char *status_line, ne_status *st)
     const char *part;
     int major, minor, status_code, klass;
 
-    /* Check they're speaking the right language */
-    status_line = strstr(status_line, "HTTP/");
-    if (status_line == NULL) {
-	return -1;
-    } 
+    /* skip leading garbage if any. */
+    part = strstr(status_line, "HTTP/");
+    if (part == NULL) return -1;
 
-    /* And find out which dialect of this peculiar language
-     * they can talk... */
-    major = 0;
-    minor = 0; 
-    /* Note, we're good children, and accept leading zero's on the
-     * version numbers */
-    for (part = status_line + 5; *part != '\0' && isdigit(*part); part++) {
+    minor = major = 0;
+
+    /* Parse version string, skipping leading zeroes. */
+    for (part += 5; *part != '\0' && isdigit(*part); part++)
 	major = major*10 + (*part-'0');
-    }
-    if (*part != '.') { 
-	return -1;
-    }
-    for (part++ ; *part != '\0' && isdigit(*part); part++) {
+
+    if (*part++ != '.') return -1;
+
+    for (;*part != '\0' && isdigit(*part); part++)
 	minor = minor*10 + (*part-'0');
-    }
-    if (*part != ' ') {
-	return -1;
-    }
+
+    if (*part != ' ') return -1;
+
     /* Skip any spaces */
-    for (; *part == ' ' ; part++) /* noop */;
-    /* Now for the Status-Code. part now points at the first Y in
-     * "HTTP/x.x YYY". We want value of YYY... could use atoi, but
-     * probably quicker this way. */
+    for (; *part == ' '; part++) /* noop */;
+
+    /* Parse the Status-Code; part now points at the first Y in
+     * "HTTP/x.x YYY". */
     if (!isdigit(part[0]) || !isdigit(part[1]) || !isdigit(part[2]) ||
-	(part[3] != '\0' && part[3] != ' ')) {
-	return -1;
-    }
+	(part[3] != '\0' && part[3] != ' ')) return -1;
     status_code = 100*(part[0]-'0') + 10*(part[1]-'0') + (part[2]-'0');
     klass = part[0]-'0';
+
     /* Skip whitespace between status-code and reason-phrase */
     for (part+=3; *part == ' ' || *part == '\t'; part++) /* noop */;
 
@@ -188,7 +227,7 @@ int ne_parse_statusline(const char *status_line, ne_status *st)
     /* Fill in the results */
     st->major_version = major;
     st->minor_version = minor;
-    st->reason_phrase = cleanse(ne_strdup(part));
+    st->reason_phrase = ne_strclean(ne_strdup(part));
     st->code = status_code;
     st->klass = klass;
     return 0;
