@@ -1,6 +1,6 @@
 /* 
-   Higher Level Interface to XML Parsers.
-   Copyright (C) 1999-2003, Joe Orton <joe@manyfish.co.uk>
+   Wrapper interface to XML parser
+   Copyright (C) 1999-2007, 2009, Joe Orton <joe@manyfish.co.uk>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -31,7 +31,7 @@
 #include <strings.h>
 #endif
 
-#include "ne_i18n.h"
+#include "ne_internal.h"
 
 #include "ne_alloc.h"
 #include "ne_xml.h"
@@ -46,12 +46,23 @@
 #include <expat.h>
 #endif
 typedef XML_Char ne_xml_char;
+
+#if !defined(XML_MAJOR_VERSION)
+#define NEED_BOM_HANDLING
+#elif XML_MAJOR_VERSION < 2 && XML_MINOR_VERSION == 95 && XML_MICRO_VERSION < 2
+#define NEED_BOM_HANDLING
+#endif
+
 #elif defined(HAVE_LIBXML)
 /* libxml2 support: */
 #include <libxml/xmlversion.h>
 #include <libxml/parser.h>
 typedef xmlChar ne_xml_char;
 
+#if LIBXML_VERSION < 20619
+/* 2.6.19 and earlier have broken BOM handling */
+#define NEED_BOM_HANDLING
+#endif
 #else /* not HAVE_LIBXML */
 #  error need an XML parser
 #endif /* not HAVE_EXPAT */
@@ -96,9 +107,13 @@ struct ne_xml_parser_s {
     struct element *current; /* current element in the branch */
     struct handler *top_handlers; /* always points at the 
 					   * handler on top of the stack. */
-    int valid; /* non-zero whilst parse should continue */
+    int failure; /* zero whilst parse should continue */
     int prune; /* if non-zero, depth within a dead branch */
 
+#ifdef NEED_BOM_HANDLING
+    int bom_pos;
+#endif
+
 #ifdef HAVE_EXPAT
     XML_Parser parser;
     char *encoding;
@@ -155,7 +170,7 @@ static xmlSAXHandler sax_handler = {
 };
 
 /* empty attributes array to mimic expat behaviour */
-static const char *empty_atts[] = {NULL, NULL};
+static const char *const empty_atts[] = {NULL, NULL};
 
 /* macro for determining the attributes array to pass */
 #define PASS_ATTS(atts) (atts ? (const char **)(atts) : empty_atts)
@@ -193,6 +208,21 @@ const char *ne_xml_doc_encoding(const ne_xml_parser *p)
 #endif
 }
 
+/* The first character of the REC-xml-names "NCName" rule excludes
+ * "Digit | '.' | '-' | '_' | CombiningChar | Extender"; the XML
+ * parser will not enforce this rule in a namespace declaration since
+ * it treats the entire attribute name as a REC-xml "Name" rule.  It's
+ * too hard to check for all of CombiningChar | Digit | Extender here,
+ * but the valid_ncname_ch1 macro catches some of the rest. */
+
+/* Return non-zero if 'ch' is an invalid start character for an NCName: */
+#define invalid_ncname_ch1(ch) ((ch) == '\0' || strchr("-.0123456789", (ch)) != NULL)
+
+/* Subversion repositories have been deployed which use property names
+ * marshalled as NCNames including a colon character; these should
+ * also be rejected but will be allowed for the time being. */
+#define invalid_ncname(xn) (invalid_ncname_ch1((xn)[0]))
+
 /* Extract the namespace prefix declarations from 'atts'. */
 static int declare_nspaces(ne_xml_parser *p, struct element *elm,
                            const ne_xml_char **atts)
@@ -200,13 +230,15 @@ static int declare_nspaces(ne_xml_parser *p, struct element *elm,
     int n;
     
     for (n = 0; atts && atts[n]; n += 2) {
-        if (strcasecmp(atts[n], "xmlns") == 0) {
+        if (strcmp(atts[n], "xmlns") == 0) {
             /* New default namespace */
             elm->default_ns = ne_strdup(atts[n+1]);
-        } else if (strncasecmp(atts[n], "xmlns:", 6) == 0) {
+        } else if (strncmp(atts[n], "xmlns:", 6) == 0) {
             struct namespace *ns;
             
-            if (atts[n][6] == '\0' || atts[n+1][0] == '\0') {
+            /* Reject some invalid NCNames as namespace prefix, and an
+             * empty URI as the namespace URI */
+            if (invalid_ncname(atts[n] + 6) || atts[n+1][0] == '\0') {
                 ne_snprintf(p->error, ERR_SIZE, 
                             ("XML parse error at line %d: invalid namespace "
                              "declaration"), ne_xml_currentline(p));
@@ -243,22 +275,20 @@ static int expand_qname(ne_xml_parser *p, struct element *elm,
         
         elm->name = ne_strdup(qname);
         elm->nspace = e->default_ns;
+    } else if (invalid_ncname(pfx + 1) || qname == pfx) {
+        ne_snprintf(p->error, ERR_SIZE, 
+                    _("XML parse error at line %d: invalid element name"), 
+                    ne_xml_currentline(p));
+        return -1;
     } else {
         const char *uri = resolve_nspace(elm, qname, pfx-qname);
 
 	if (uri) {
-	    /* The name is everything after the ':' */
-	    if (pfx[1] == '\0') {
-		ne_snprintf(p->error, ERR_SIZE, 
-			    ("XML parse error at line %d: element name missing"
-                             "after namespace prefix"), ne_xml_currentline(p));
-		return -1;
-	    }
 	    elm->name = ne_strdup(pfx+1);
             elm->nspace = uri;
 	} else {
 	    ne_snprintf(p->error, ERR_SIZE, 
-                        ("XML parse error at line %d: undeclared namespace"),
+                        ("XML parse error at line %d: undeclared namespace prefix"),
                         ne_xml_currentline(p));
 	    return -1;
 	}
@@ -275,7 +305,7 @@ static void start_element(void *userdata, const ne_xml_char *name,
     struct handler *hand;
     int state = NE_XML_DECLINE;
 
-    if (!p->valid) return;
+    if (p->failure) return;
     
     if (p->prune) {
         p->prune++;
@@ -288,7 +318,7 @@ static void start_element(void *userdata, const ne_xml_char *name,
     p->current = elm;
 
     if (declare_nspaces(p, elm, atts) || expand_qname(p, elm, name)) {
-        p->valid = 0;
+        p->failure = 1;
         return;
     }
 
@@ -300,7 +330,7 @@ static void start_element(void *userdata, const ne_xml_char *name,
                                   elm->nspace, elm->name, PASS_ATTS(atts));
     }
 
-    NE_DEBUG(NE_DBG_XMLPARSE, "XML: start-element (%d, {%s, %s}) => %d\n", 
+    NE_DEBUG(NE_DBG_XML, "XML: start-element (%d, {%s, %s}) => %d\n", 
              elm->parent->state, elm->nspace, elm->name, state);             
     
     if (state > 0)
@@ -308,8 +338,8 @@ static void start_element(void *userdata, const ne_xml_char *name,
     else if (state == NE_XML_DECLINE)
         /* prune this branch. */
         p->prune++;
-    else /* state == NE_XML_ABORT */
-        p->valid = 0;
+    else /* state < 0 => abort parse  */
+        p->failure = state;
 }
 
 /* Destroys an element structure. */
@@ -325,7 +355,7 @@ static void destroy_element(struct element *elm)
 	ne_free(this_ns->uri);
 	ne_free(this_ns);
 	this_ns = next_ns;
-    };
+    }
     if (elm->default_ns)
         ne_free(elm->default_ns);
     ne_free(elm);
@@ -337,12 +367,12 @@ static void char_data(void *userdata, const ne_xml_char *data, int len)
     ne_xml_parser *p = userdata;
     struct element *elm = p->current;
 
-    if (!p->valid || p->prune) return;
+    if (p->failure || p->prune) return;
     
-    if (elm->handler->cdata_cb && 
-        elm->handler->cdata_cb(elm->handler->userdata, elm->state, data, len)) {
-        NE_DEBUG(NE_DBG_XML, "Cdata callback failed.\n");
-        p->valid = 0;
+    if (elm->handler->cdata_cb) {
+        p->failure = elm->handler->cdata_cb(elm->handler->userdata, elm->state, data, len);
+        NE_DEBUG(NE_DBG_XML, "XML: char-data (%d) returns %d\n", 
+                 elm->state, p->failure);
     }        
 }
 
@@ -352,18 +382,20 @@ static void end_element(void *userdata, const ne_xml_char *name)
     ne_xml_parser *p = userdata;
     struct element *elm = p->current;
 
-    if (!p->valid) return;
+    if (p->failure) return;
 	
     if (p->prune) {
         if (p->prune-- > 1) return;
-    } else if (elm->handler->endelm_cb && 
-               elm->handler->endelm_cb(elm->handler->userdata, elm->state,
-                                       elm->nspace, elm->name)) {
-        NE_DEBUG(NE_DBG_XML, "XML: end-element for %d failed.\n", elm->state);
-        p->valid = 0;
+    } else if (elm->handler->endelm_cb) {
+        p->failure = elm->handler->endelm_cb(elm->handler->userdata, elm->state,
+                                             elm->nspace, elm->name);
+        if (p->failure) {
+            NE_DEBUG(NE_DBG_XML, "XML: end-element for %d failed with %d.\n", 
+                     elm->state, p->failure);
+        }
     }
     
-    NE_DEBUG(NE_DBG_XMLPARSE, "XML: end-element (%d, {%s, %s})\n",
+    NE_DEBUG(NE_DBG_XML, "XML: end-element (%d, {%s, %s})\n",
              elm->state, elm->nspace, elm->name);
 
     /* move back up the tree */
@@ -373,6 +405,28 @@ static void end_element(void *userdata, const ne_xml_char *name)
     destroy_element(elm);
 }
 
+#if defined(HAVE_EXPAT) && XML_MAJOR_VERSION > 1
+/* Stop the parser if an entity declaration is hit. */
+static void entity_declaration(void *userData, const XML_Char *entityName,
+                              int is_parameter_entity, const XML_Char *value,
+                              int value_length, const XML_Char *base,
+                              const XML_Char *systemId, const XML_Char *publicId,
+                              const XML_Char *notationName)
+{
+    ne_xml_parser *parser = userData;
+    
+    NE_DEBUG(NE_DBG_XMLPARSE, "XML: entity declaration [%s]. Failing.\n",
+             entityName);
+
+    XML_StopParser(parser->parser, XML_FALSE);
+}
+#elif defined(HAVE_EXPAT)
+/* A noop default_handler. */
+static void default_handler(void *userData, const XML_Char *s, int len)
+{
+}
+#endif
+
 /* Find a namespace definition for 'prefix' in given element, where
  * length of prefix is 'pfxlen'.  Returns the URI or NULL. */
 static const char *resolve_nspace(const struct element *elm, 
@@ -394,11 +448,25 @@ static const char *resolve_nspace(const struct element *elm,
     return NULL;
 }
 
+const char *ne_xml_resolve_nspace(ne_xml_parser *parser, 
+                                  const char *prefix, size_t length)
+{
+    if (prefix) {
+        return resolve_nspace(parser->current, prefix, length);
+    }
+    else {
+        struct element *e = parser->current;
+
+        while (e->default_ns == NULL)
+            e = e->parent;
+
+        return e->default_ns;
+    }
+}
+
 ne_xml_parser *ne_xml_create(void) 
 {
     ne_xml_parser *p = ne_calloc(sizeof *p);
-    /* Initialize other stuff */
-    p->valid = 1;
     /* Placeholder for the root element */
     p->current = p->root = ne_calloc(sizeof *p->root);
     p->root->default_ns = "";
@@ -413,14 +481,34 @@ ne_xml_parser *ne_xml_create(void)
     XML_SetCharacterDataHandler(p->parser, char_data);
     XML_SetUserData(p->parser, (void *) p);
     XML_SetXmlDeclHandler(p->parser, decl_handler);
+
+    /* Prevent the "billion laughs" attack against expat by disabling
+     * internal entity expansion.  With 2.x, forcibly stop the parser
+     * if an entity is declared - this is safer and a more obvious
+     * failure mode.  With older versions, installing a noop
+     * DefaultHandler means that internal entities will be expanded as
+     * the empty string, which is also sufficient to prevent the
+     * attack. */
+#if XML_MAJOR_VERSION > 1
+    XML_SetEntityDeclHandler(p->parser, entity_declaration);
 #else
+    XML_SetDefaultHandler(p->parser, default_handler);
+#endif
+
+#else /* HAVE_LIBXML */
     p->parser = xmlCreatePushParserCtxt(&sax_handler, 
 					(void *)p, NULL, 0, NULL);
     if (p->parser == NULL) {
 	abort();
     }
+#if LIBXML_VERSION < 20602
     p->parser->replaceEntities = 1;
+#else
+    /* Enable expansion of entities, and disable network access. */
+    xmlCtxtUseOptions(p->parser, XML_PARSE_NOENT | XML_PARSE_NONET);
 #endif
+
+#endif /* HAVE_LIBXML || HAVE_EXPAT */
     return p;
 }
 
@@ -448,62 +536,83 @@ void ne_xml_push_handler(ne_xml_parser *p,
     }
 }
 
-void ne_xml_parse_v(void *userdata, const char *block, size_t len) 
+int ne_xml_parse_v(void *userdata, const char *block, size_t len) 
 {
     ne_xml_parser *p = userdata;
-    /* FIXME: The two XML parsers break all our nice abstraction by
-     * choosing different char *'s. The swine. This cast will come
-     * back and bite us someday, no doubt. */
-    ne_xml_parse(p, block, len);
+    return ne_xml_parse(p, (const ne_xml_char *)block, len);
 }
 
-/* Parse the given block of input of length len */
-void ne_xml_parse(ne_xml_parser *p, const char *block, size_t len) 
+#define BOM_UTF8 "\xEF\xBB\xBF" /* UTF-8 BOM */
+
+int ne_xml_parse(ne_xml_parser *p, const char *block, size_t len) 
 {
     int ret, flag;
     /* duck out if it's broken */
-    if (!p->valid) {
-	NE_DEBUG(NE_DBG_XML, "Not parsing %" NE_FMT_SIZE_T " bytes.\n", 
-		 len);
-	return;
+    if (p->failure) {
+	NE_DEBUG(NE_DBG_XMLPARSE, "XML: Failed; ignoring %" NE_FMT_SIZE_T 
+                 " bytes.\n", len);
+	return p->failure;
     }
     if (len == 0) {
 	flag = -1;
 	block = "";
-	NE_DEBUG(NE_DBG_XML, "Got 0-length buffer, end of document.\n");
+	NE_DEBUG(NE_DBG_XMLPARSE, "XML: End of document.\n");
     } else {	
-	NE_DEBUG(NE_DBG_XML, "Parsing %" NE_FMT_SIZE_T " length buffer.\n",
-		 len);
+	NE_DEBUG(NE_DBG_XMLPARSE, "XML: Parsing %" NE_FMT_SIZE_T " bytes.\n", len);
 	flag = 0;
     }
-    /* Note, don't write a parser error if !p->valid, since an error
+
+#ifdef NEED_BOM_HANDLING
+    if (p->bom_pos < 3) {
+        NE_DEBUG(NE_DBG_XMLPARSE, "Checking for UTF-8 BOM.\n");
+        while (len > 0 && p->bom_pos < 3 && 
+               block[0] == BOM_UTF8[p->bom_pos]) {
+            block++;
+            len--;
+            p->bom_pos++;
+        }
+        if (len == 0)
+            return 0;
+        if (p->bom_pos == 0) {
+            p->bom_pos = 3; /* no BOM */
+        } else if (p->bom_pos > 0 && p->bom_pos < 3) {
+            ne_strnzcpy(p->error, _("Invalid Byte Order Mark"), sizeof p->error);
+            return p->failure = 1;
+        }
+    }
+#endif
+
+    /* Note, don't write a parser error if p->failure, since an error
      * will already have been written in that case. */
 #ifdef HAVE_EXPAT
     ret = XML_Parse(p->parser, block, len, flag);
-    NE_DEBUG(NE_DBG_XMLPARSE, "XML_Parse returned %d\n", ret);
-    if (ret == 0 && p->valid) {
+    NE_DEBUG(NE_DBG_XMLPARSE, "XML: XML_Parse returned %d\n", ret);
+    if (ret == 0 && p->failure == 0) {
 	ne_snprintf(p->error, ERR_SIZE,
-		    "XML parse error at line %d: %s", 
+		    "XML parse error at line %" NE_FMT_XML_SIZE ": %s", 
 		    XML_GetCurrentLineNumber(p->parser),
 		    XML_ErrorString(XML_GetErrorCode(p->parser)));
-	p->valid = 0;
+	p->failure = 1;
+        NE_DEBUG(NE_DBG_XMLPARSE, "XML: Parse error: %s\n", p->error);
     }
 #else
     ret = xmlParseChunk(p->parser, block, len, flag);
-    NE_DEBUG(NE_DBG_XMLPARSE, "xmlParseChunk returned %d\n", ret);
+    NE_DEBUG(NE_DBG_XMLPARSE, "XML: xmlParseChunk returned %d\n", ret);
     /* Parse errors are normally caught by the sax_error() callback,
      * which clears p->valid. */
-    if (p->parser->errNo && p->valid) {
-	ne_snprintf(p->error, ERR_SIZE, "XML parse error at line %d.", 
+    if (p->parser->errNo && p->failure == 0) {
+	ne_snprintf(p->error, ERR_SIZE, "XML parse error at line %d", 
 		    ne_xml_currentline(p));
-	p->valid = 0;
+	p->failure = 1;
+        NE_DEBUG(NE_DBG_XMLPARSE, "XML: Parse error: %s\n", p->error);
     }
 #endif
+    return p->failure;
 }
 
-int ne_xml_valid(ne_xml_parser *p)
+int ne_xml_failed(ne_xml_parser *p)
 {
-    return p->valid;
+    return p->failure;
 }
 
 void ne_xml_destroy(ne_xml_parser *p) 
@@ -539,7 +648,7 @@ void ne_xml_destroy(ne_xml_parser *p)
 
 void ne_xml_set_error(ne_xml_parser *p, const char *msg)
 {
-    ne_snprintf(p->error, ERR_SIZE, msg);
+    ne_snprintf(p->error, ERR_SIZE, "%s", msg);
 }
 
 #ifdef HAVE_LIBXML
@@ -553,11 +662,12 @@ static void sax_error(void *ctx, const char *msg, ...)
     ne_vsnprintf(buf, 1024, msg, ap);
     va_end(ap);
 
-    ne_snprintf(p->error, ERR_SIZE, 
-		_("XML parse error at line %d: %s."),
-		p->parser->input->line, buf);
-
-    p->valid = 0;
+    if (p->failure == 0) {
+        ne_snprintf(p->error, ERR_SIZE, 
+                    _("XML parse error at line %d: %s"),
+                    p->parser->input->line, buf);
+        p->failure = 1;
+    }
 }
 #endif
 
